#!/usr/bin/env python
# -*- coding: utf-8 -*-

VERSION = "0.4.0"

# available languages :
LANGUAGES = {
    'en': 'English',
    'es': 'Español',
    'fr': 'Français',
}

DEFAULT_LOGIN = "zenfeed"
DEFAULT_PASSWORD = "zenfeed"
DEFAULT_REFRESH_INTERVAL = 3600
DEFAULT_MAX_ENTRIES = 1000
DEFAULT_ENTRIES_PER_PAGE = 50
DEFAULT_HIGHLIGHT_NEWS = True



#!/usr/bin/env python
# -*- coding:utf-8 -*-

from __future__ import print_function, unicode_literals
import unittest

import feedparser
from path import path

from sys import version_info as python_version
if python_version[0] == 3:
    basestring = (bytes, str)

class TestFeedParsing(unittest.TestCase):

    def setUp(self):
        self.files = path("test/fixtures/feeds").listdir()

    def test_feeds_parsing_succeed(self):
        self.assertTrue(len(self.files) > 0)
        for filename in self.files:
            feed = feedparser.parse(filename)
            self.assertIsInstance(feed, feedparser.FeedParserDict, msg=filename)


class TestFeedFields(unittest.TestCase):

    def setUp(self):
        files = path("test/fixtures/feeds").listdir()
        self.feeds = [feedparser.parse(filename) for filename in files]

    def test_feeds_have_string_title(self):
        for feed in self.feeds:
            self.assertIsInstance(feed['feed']['title'], basestring)


if __name__ == '__main__':
    unittest.main()




#!/usr/bin/env python
# -*- coding:utf-8 -*-

"""Zen RSS feed reader.

Usage:
  zenfeed [--database URI --favicons PATH -p PORT --log LOG --lang LANG
           --tz TIMEZONE --prefix PREFIX --no-cache --debug]
  zenfeed genstatic PATH
  zenfeed -h | --help
  zenfeed -v | --version

Options:
  -h --help           Show this screen.
  -v --version        Show version.
  -d --database URI   Specify database URI (SQLAlchemy format).
                      For SQLite, this is just a file name.
                      [default: ./zenfeed.db]
  -f --favicons PATH  Specify where favicons will be put.
                      [default: ./]
  -p --port PORT      Specify on which port to listen.
                      [default: 5000]
  --prefix PREFIX     If zenfeed is not alone in its domain/subdomain,
                      specify its path prefix. E.g: /zenfeed/
                      [default: /]
  --log LOG           Specify where to log messages, and which level to set.
                      Can be "stderr", "syslog", or a filename, followed by the level.
                      [default: stderr:INFO]
  --lang LANG         Fix the language instead of let it depend on the browser's value.
                      The language needs to be supported. E.g: en
                      [default: browser]
  --tz TIMEZONE       Specify which timezone to use, to adjust date and time display.
                      For french timezone : "--tz Europe/Paris"
                      [default: GMT]
  --no-cache          Disable cache on index and feed pages.
  --debug             Debug mode, do not use.

"""

from __future__ import unicode_literals, print_function#, absolute_import
from docopt import docopt
from os import urandom
from path import path
from pytz import all_timezones
import logging

# avoid side effects with imports…
import gevent
from gevent.monkey import patch_socket, patch_ssl

from caching import Cache
from log import setup_logger, logger
from settings import LANGUAGES, VERSION
try:
    import zenfeed
except ImportError:
    class Obj: pass
    zenfeed = Obj()
    zenfeed.__path__ = ['./']

def genstatic(dst):
    path.copytree(path(zenfeed.__path__[0]) / 'static', dst)

def main():
    args = docopt(__doc__, version='zenfeed ' + VERSION)

    log_arg, log_level = args['--log'].rsplit(':', 1)
    if log_arg not in ('stderr', 'syslog'):
        setup_logger(type='file', filename=path(log_arg).abspath(),
                     level=log_level)
    else:
        setup_logger(type=log_arg, level=log_level)

    logger.info('Zenfeed %s booting...', VERSION)

    if args['genstatic']:
        return genstatic(args['PATH'])

    port = int(args['--port'])

    cache_disabled = args['--no-cache']

    path_prefix = args['--prefix']
    if path_prefix.endswith('/'):
        path_prefix = path_prefix[:-1]
    if path_prefix and not path_prefix.startswith('/'):
        path_prefix = '/' + path_prefix

    fixed_language = args['--lang']
    if fixed_language == 'browser':
        fixed_language = None
    else:
        logger.info('Language fixed to "%s"', fixed_language)
        if (fixed_language not in LANGUAGES
            and fixed_language.split('_', 1)[0] not in LANGUAGES):
            return logger.critical('Fixed language not supported !')

    fixed_timezone = args['--tz']
    logger.info('Timezone fixed to "%s"', fixed_timezone)
    if fixed_timezone not in all_timezones:
        return logger.critical('Fixed timezone not supported !')

    db_uri = args['--database']
    if db_uri == ':memory:':
        db_uri = 'sqlite://'
    elif not "://" in db_uri:
        db_uri = 'sqlite:///%s' % path(db_uri).abspath()

    import app as app_module
    app = app_module.create_flask_app(prefix=path_prefix)
    app.config.update(
        DEBUG = args['--debug'],
        SQL_DEBUG = False,
        SECRET_KEY = urandom(32),
        SQLALCHEMY_DATABASE_URI = db_uri,
        FAVICON_DIR = path(args['--favicons']).abspath(),
        FIXED_LANGUAGE = fixed_language,
        FIXED_TIMEZONE = fixed_timezone,
        CACHE_ENABLED = not cache_disabled,
        PATH_PREFIX = path_prefix,
    )
    Cache(app)

    from models import setup_tables, Feed
    patch_socket()
    patch_ssl()
    setup_tables()

    from deadline_manager import deadlineManager
    import views
    from werkzeug.contrib.fixers import ProxyFix
    app.wsgi_app = ProxyFix(app.wsgi_app)

    feeds = Feed.query.all()
    deadlineManager.favicon_dir = path(args['--favicons']).abspath()
    deadlineManager.launch_deadline_workers(feeds)
    deadlineManager.start()

    logger.info("Server started at port %d (prefix: %s/)", port, path_prefix)
    if args['--debug']:
        logger.warning("DEBUG mode activated")
        app.run(host='0.0.0.0', port=port, debug=True)
    else:
        from gevent.wsgi import WSGIServer
        http_server = WSGIServer(('0.0.0.0', port), app)
        try:
            http_server.serve_forever()
        except KeyboardInterrupt:
            pass
    #gevent.joinall([deadlineManager])

if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding:utf-8 -*-

import logging
from logging import handlers
logger = logging.getLogger('zenfeed')


def parse_level(level):
    return {
        'debug': logging.DEBUG,
        'info': logging.INFO,
        'warn': logging.WARNING,
        'warning': logging.WARNING,
        'error': logging.ERROR,
        'critical': logging.CRITICAL,
    }[level]

def setup_logger(**params):
    params['level'] = parse_level(params['level'].lower())
    if not 'format' in params:
        params['format'] = "%(asctime)s %(name)-8s %(levelname)-8s | %(message)s"
    if params['type'] == 'file':
        handler = logging.FileHandler(params['filename'])
    elif params['type'] == 'stderr':
        handler = logging.StreamHandler()
    elif params['type'] == 'syslog':
        handler = handlers.SysLogHandler(address='/dev/log')
    logger.setLevel(params['level'])
    handler.setLevel(params['level'])
    handler.setFormatter(logging.Formatter(params['format']))
    logger.addHandler(handler)



#!/usr/bin/env python
# -*- coding:utf-8 -*-

from __future__ import unicode_literals, print_function

from functools import wraps

class Cache(object):
	def __init__(self, app):
		self.memory = {}
		self.enabled = app.config['CACHE_ENABLED']
		app.cache = self

	def to_key(self, args, kwargs):
		tup1 = tuple(args)
		tup2 = tuple(kwargs.items())
		return (tup1, tup2)

	def get(self, *args, **kwargs):
		if not self.enabled: return
		key = self.to_key(args, kwargs)
		return self.memory[key]

	def put(self, value, *args, **kwargs):
		if not self.enabled: return value
		key = self.to_key(args, kwargs)
		self.memory[key] = value
		return value

	def has(self, *args, **kwargs):
		if not self.enabled: return
		key = self.to_key(args, kwargs)
		return key in self.memory

	def delete(self, *args, **kwargs):
		if not self.enabled: return
		key = self.to_key(args, kwargs)
		if key in self.memory:
			del self.memory[key]



#!/usr/bin/env python
# -*- coding:utf-8 -*-

from __future__ import unicode_literals, print_function
from hashlib import sha256
from path import path
from bs4 import BeautifulSoup
import requests
import feedparser

from log import logger

class FaviconException(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)

class FetchingException(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)

# Utils :

def sanitize_url(url):
    if not (url.startswith('http://') or url.startswith('https://')):
        url = 'http://%s' % url
    return url

def parse_url(url):
    url = sanitize_url(url)
    # delete eventual query string and anchor:
    url_begin = url.partition('#')[0].partition('?')[0]
    url_host = '/'.join(url_begin.split('/')[:3])
    return url_begin, url_host

def concat_urls(url, path):
    url_begin, url_host = parse_url(url)
    if path.startswith('/'):
        return url_host + path
    else:
        if not url_begin.endswith('/'):
            url_begin = url_begin + '/'
        return url_begin + path

def fetch_url(url):
    url = sanitize_url(url)
    resp = requests.get(url, verify=False)
    return resp


# Feed stuff :

def fetch_and_parse_feed(url, etag=None, last_modified=None):
    # TODO implement etag & last_modified header
    url = sanitize_url(url)
    feed_parsed = feedparser.parse(url)
    if not hasattr(feed_parsed, 'status'):
        raise FetchingException("Connection error")
    elif feed_parsed.status not in (200, 301, 302):
        raise FetchingException("status_code is %d" % feed_parsed.status)
    if feed_parsed.version == '':
        # it's probably html instead of rss/atom
        resp = fetch_url(url)
        if resp.status_code not in (200, 301, 302):
            raise FetchingException("status_code is %d" % resp.status_code)
        soup = BeautifulSoup(resp.content)
        try:
            url = soup.find_all("link", rel="alternate")[0]['href']
        except (IndexError, KeyError):
            # alternate-link is missing
            raise FetchingException("Neither RSS nor good HTML...")
        if not url.startswith("http"):
            url = concat_urls(resp.url, url)
        feed_parsed = feedparser.parse(url)
        if feed_parsed.status not in (200, 301, 302):
            raise FetchingException("status_code is %d" % feed_parsed.status)
    if feed_parsed.status == 302:  # moved permanently
        logger.warning("/!\\ permanent redirect (302) for %s", url)
        url = feed_parsed.href
    elif feed_parsed.status == 301:
        logger.warning("/!\\ temporary redirect (301) for %s", url)
    return {"feed": feed_parsed, "real_url": url}


# Favicon stuff :

def fetch_icon_in_page(resp):
    if resp.status_code != 200:
        raise FaviconException("Http code != 200 for url: " + resp.url)
    soup = BeautifulSoup(resp.content)
    icon_tags = [tag for tag in soup.find_all("link")
                 if tag.has_attr("rel")
                 and "icon" in tag.attrs["rel"]
                 and tag.has_attr("href")]
    if not icon_tags:
        raise FaviconException("No icon tags found in: " + resp.url)
    icon = icon_tags[0]
    icon_href = icon.attrs["href"]

    if icon_href.startswith("http"):
        icon_url = icon_href
    else:
        icon_url = concat_urls(resp.url, icon_href)
    return icon_url

def fetch_icon_url(url):
    resp = fetch_url(url)
    return fetch_icon_in_page(resp)

def fetch_root_icon_url(url):
    _, root_url = parse_url(url)
    return fetch_icon_url(root_url)

def default_favicon_url(url):
    return concat_urls(url, '/favicon.ico')

def fetch_favicon(url, page=None):
    try:
        # search icon url in page:
        if page is None:
            icon_url = fetch_icon_url(url)
        else:
            icon_url = fetch_icon_in_page(page)
    except FaviconException:
        try:
            # search icon url in root page ('/'):
            icon_url = fetch_root_icon_url(url)
        except FaviconException:
            # fallback to /favicon.ico:
            icon_url = default_favicon_url(url)
    resp = fetch_url(icon_url)
    return resp

def mocked_default_favicon():
    class Obj: pass
    obj = Obj()
    obj.url = 'default'
    obj.content = (path(__file__).dirname().relpath() / 'static' / 'img' / 'feedicon.png').bytes()
    return obj

def save_favicon(resp, directory):
    if resp.status_code != 200 or len(resp.content) < 10:
        resp = mocked_default_favicon()
    name = sha256(resp.url.encode('utf-8')).hexdigest()
    save_image(resp.content, path(directory) / name)
    return name

def save_image(content, name):
    with open(name, 'wb') as f:
        f.write(content)



#!/usr/bin/env python
# -*- coding:utf-8 -*-

from __future__ import unicode_literals, print_function
import gevent
from gevent.queue import Queue, Empty
from requests.exceptions import ConnectionError

from actor import Mail
from builder import FeedFromDict, EntryFromDict
from fetcher import (fetch_and_parse_feed, sanitize_url,
                     save_favicon, fetch_favicon, FetchingException)
from log import logger
from app import app
from flask import url_for
from models import db, Feed, Entry, Config, update_feed, create_or_update_entry


def new_feed_worker(url, favicon_dir, answer_box, manager_box):
    logger.info("Fetching feed... [%s]", url)
    try:
        fetched = fetch_and_parse_feed(url)
        feed_dict, real_url = fetched['feed'], fetched['real_url']
    except FetchingException:
        return answer_box.put(Exception("Error with feed: " + url))

    feed = FeedFromDict(feed_dict, Config.get())
    feed.url = sanitize_url(real_url) # set the real feed url
    logger.info("Fetching favicon... [%s]", feed.url)
    feed.favicon_path = save_favicon(fetch_favicon(feed.url), favicon_dir)
    db.session.add(feed)
    for e in feed_dict['entries'][::-1]:
        entry = EntryFromDict(e, feed.url)
        entry.feed = feed # set the corresponding feed
        db.session.add(entry)
    db.session.commit()

    most_recent_entry = feed.entries.order_by(Entry.updated.desc()).first()
    feed.updated = most_recent_entry.updated
    db.session.commit()

    answer_box.put(feed)
    manager_box.put({'type': 'new-deadline-worker', 'feed_id': feed.id})
    manager_box.put({'type': 'refresh-cache', 'feed_id': feed.id})


def fetching_work(feed, manager_box):
    fetched = fetch_and_parse_feed(feed.url)
    feed_dict, real_url = fetched['feed'], fetched['real_url']
    if feed.url != real_url:
        logger.warning("©©© Feed url changed from %s to %s", feed.url, real_url)
        feed.url = sanitize_url(real_url)
    logger.info("©©© Updated feed: %s", feed.url)
    feed_changed = update_feed(feed, FeedFromDict(feed_dict, Config.get()))
    any_entry_created = any([
        create_or_update_entry(feed.id, EntryFromDict(e, feed.url))
        for e in feed_dict['entries']
    ])
    db.session.commit()

    if any_entry_created:
        most_recent_entry = feed.entries.order_by(Entry.updated.desc()).first()
        feed.updated = most_recent_entry.updated
        feed.has_news = feed.has_news or feed.highlight_news
        db.session.merge(feed)
        db.session.commit()
        manager_box.put({'type': 'refresh-cache', 'feed_id': feed.id})


def deadline_worker(feed, inbox, manager_box):
    while True:
        try:
            msg = inbox.get(timeout=feed.refresh_interval)
        except Empty:
            msg = None # timeout -> refresh !
        feed = Feed.query.get(feed.id)  # refresh
        if msg is not None:
            # msg types :
            # - stop
            # - refresh
            if isinstance(msg, Mail):
                if msg.body['type'] == 'stop':
                    return msg.answer_box.put('okay')
            logger.warning("©©© Refresh forced.")
        try:
            fetching_work(feed, manager_box)
        except FetchingException as e:
            logger.error("Error re-fetching feed %s : %s", feed.url, e.value)
            continue
        except ConnectionError:
            logger.error("Connection error re-fetching feed %s", feed.url)
            continue


def cache_worker(feed_id):
    with app.test_request_context('/'):
        if feed_id is not None:
            app.cache.delete(feed_id=feed_id)
            app.view_functions['feed_view'](feed_id=feed_id, bot_flag=True)
        app.cache.delete('/')
        app.view_functions['index']()


def delete_worker(worker, feed_id, answer_box, kill_timeout=30):
    if worker is not None:
        logger.warning("Going to stop gracefully feed %d", feed_id)
        inbox = Queue()
        mail = Mail(inbox, {'type': 'stop'})
        worker['queue'].put(mail)
        resp = inbox.get(timeout=kill_timeout)
        if resp is None:
            logger.warning("Going to KILL feed %d !", feed_id)
            worker['worker'].kill()
    feed = Feed.query.get(feed_id)
    if feed is None:
        logger.error("Why doesn't feed [%d] exist ?!", feed_id)
        return answer_box.put({'success': False})
    db.session.delete(feed)
    db.session.commit()
    logger.warning("Okay, feed %d has been deleted !", feed_id)
    answer_box.put({'success': True})



#!/usr/bin/env python
# -*- coding:utf-8 -*-

from __future__ import print_function, unicode_literals
import unittest

import feedparser
from path import path
from zenfeed.builder import normalize_feed_dict, normalize_entry_dict

from sys import version_info as python_version
if python_version[0] == 3:
    basestring = (bytes, str)


class TestBuilding(unittest.TestCase):

    def setUp(self):
        files = path("test/fixtures/feeds").listdir()
        self.feeds = [feedparser.parse(filename) for filename in files]

    def test_feeds_have_string_url(self):
        for feed in self.feeds:
            dico = normalize_feed_dict(feed)
            url_begin = dico['url'][:4]
            self.assertEqual(url_begin, 'http')

    def test_entries_have_string_url(self):
        for feed in self.feeds:
            entries = feed['entries']
            for e in entries:
                dico = normalize_entry_dict(e)
                self.assertIsInstance(dico['content'], basestring, msg=dico['url'])
                url_begin = dico['url'][:4]
                self.assertEqual(url_begin, 'http')


if __name__ == '__main__':
    unittest.main()




#!/usr/bin/env python
# -*- coding:utf-8 -*-

from __future__ import unicode_literals, print_function
from hashlib import sha256
from flask.ext.sqlalchemy import SQLAlchemy
from sqlalchemy.orm import class_mapper, ColumnProperty

from app import app
from log import logger
from migrations import run_migrations
from settings import (VERSION, DEFAULT_LOGIN, DEFAULT_PASSWORD,
    DEFAULT_REFRESH_INTERVAL, DEFAULT_MAX_ENTRIES, DEFAULT_HIGHLIGHT_NEWS,
    DEFAULT_ENTRIES_PER_PAGE)


db = SQLAlchemy(app)

def setup_tables():
    db.create_all()
    if Config.get() is None:
        logger.info("Creating Config single instance")
        Config.query.delete()
        c = Config(
            version=VERSION,
            login=DEFAULT_LOGIN,
            password=Config.hash(DEFAULT_PASSWORD),
            default_refresh_interval=DEFAULT_REFRESH_INTERVAL,
            default_max_entries=DEFAULT_MAX_ENTRIES,
            default_highlight_news=DEFAULT_HIGHLIGHT_NEWS,
            default_entries_per_page=DEFAULT_ENTRIES_PER_PAGE,
        )
        db.session.add(c)
        db.session.commit()
        Config.refresh_instance()
    elif Config.get().version != VERSION:
        run_migrations(from_version=Config.get().version)
    db.engine.echo = app.config['SQL_DEBUG']


tags = db.Table('tags',
    db.Column('tag_id', db.Integer, db.ForeignKey('tag.id')),
    db.Column('feed_id', db.Integer, db.ForeignKey('feed.id'))
)


class Base(object):
    @classmethod
    def fields_set(cls, exclude_fields=None):
        exclude_fields = exclude_fields or []
        my_fields = [prop.key for prop in class_mapper(cls).iterate_properties
                     if isinstance(prop, ColumnProperty)]
        return set(my_fields) - set(exclude_fields)

    def to_dict(self, exclude_fields=None):
        fields = self.fields_set(exclude_fields)
        return dict((field, getattr(self, field)) for field in fields)



class Tag(db.Model, Base):
    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(50))

    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return u'<Tag %r>' % self.name


class Feed(db.Model, Base):
    id = db.Column(db.Integer, primary_key=True)
    url = db.Column(db.String, nullable=False, unique=True)
    entries = db.relationship('Entry', backref='feed', lazy='dynamic')
    tags = db.relationship('Tag', secondary=tags,
        backref=db.backref('feeds', lazy='dynamic'), lazy='dynamic')

    title = db.Column(db.String)
    link = db.Column(db.String)
    subtitle = db.Column(db.String)
    author = db.Column(db.String)
    generator = db.Column(db.String)
    encoding = db.Column(db.String)
    updated = db.Column(db.DateTime)
    entries_hash = db.Column(db.String) # last updated entries hash
    refresh_interval = db.Column(db.Integer)
    max_entries = db.Column(db.Integer)
    entries_per_page = db.Column(db.Integer)
    highlight_news = db.Column(db.Boolean)
    has_news = db.Column(db.Boolean)
    favicon_path = db.Column(db.String)

    def __init__(self, url):
        self.url = url

    def __repr__(self):
        return u'<Feed %r>' % self.url

    def last_entry(self):
        return self.entries.order_by(Entry.updated.desc()).first()


class Entry(db.Model, Base):
    id = db.Column(db.Integer, primary_key=True)
    feed_id = db.Column(db.Integer, db.ForeignKey('feed.id'))

    url = db.Column(db.String) # ['id']
    link = db.Column(db.String) # ['link']
    title = db.Column(db.String)
    content = db.Column(db.Text) # ['value'], or ['content'][0]['value']
    mimetype = db.Column(db.String) # ['value'], or ['content'][0]['type']
    created = db.Column(db.DateTime) # or published
    updated = db.Column(db.DateTime)
    public = db.Column(db.Boolean)

    def __init__(self, title, content):
        self.title = title
        self.content = content

    def __repr__(self):
        return u'<Entry %r>' % self.title


class Config(db.Model, Base):
    id = db.Column(db.Integer, primary_key=True)
    version = db.Column(db.String)
    login = db.Column(db.String, nullable=False, unique=True)
    password = db.Column(db.String, nullable=False, unique=True)

    default_refresh_interval = db.Column(db.Integer)
    default_max_entries = db.Column(db.Integer)
    default_entries_per_page = db.Column(db.Integer)
    default_highlight_news = db.Column(db.Boolean)

    instance = None
    salt = "saltpepperandcurry"

    @staticmethod
    def get():
        if Config.instance is None:
            Config.refresh_instance()
        return Config.instance

    @staticmethod
    def refresh_instance():
        Config.instance = Config.query.get(1)

    @staticmethod
    def validate_password(pw):
        return Config.get().password == pw

    @staticmethod
    def change_password(pw):
        logger.warning("Password changed !")
        Config.get().password = Config.hash(pw)

    @staticmethod
    def hash(pw):
        string = (pw + Config.salt).encode('utf-8')
        return sha256(string).hexdigest()

    def __repr__(self):
        return u'<Config of %r>' % self.login


def _update_attr(attr_name, old_obj, new_obj):
    """ Update an object if necessary. Return True if updated. """

    old_value = getattr(old_obj, attr_name)
    new_value = getattr(new_obj, attr_name)
    if new_value != old_value and new_value is not None:
        setattr(old_obj, attr_name, new_value)
        return True


def update_feed(feed, new_feed):
    """ Return True if the feed has changed. """
    assert isinstance(feed, Feed)
    assert isinstance(new_feed, Feed)
    fields = ['title', 'link', 'subtitle', 'author', 'generator',
              'encoding',] # no more ['updated', 'entries_hash'] fields

    changed = any([_update_attr(f, feed, new_feed) for f in fields])
    if changed:
        db.session.merge(feed)
    return changed


def create_or_update_entry(feed_id, new_entry):
    """ Return True if the entry is new.
        When updated, I assume the 'updated' field has changed.
        (I don't force this behaviour). """
    # assert isinstance(feed, Feed)
    assert isinstance(new_entry, Entry)
    fields = ['url', 'link', 'title', 'content', 'mimetype',
              'created', 'updated']

    entry = Entry.query.filter_by(feed_id=feed_id, url=new_entry.url).first()
    if entry is None: # create
        new_entry.feed_id = feed_id
        logger.info("→→ new entry: %s", new_entry.title)
        db.session.merge(new_entry)
        return True
    else: # update
        changed_fields = [_update_attr(f, entry, new_entry) for f in fields]
        changed_fields_strings = [s for s,b in zip(fields, changed_fields) if b]
        changed = any(changed_fields)
        if changed:
            logger.info("→→ changed entry: %s | %s", entry.title, repr(changed_fields_strings))
            db.session.merge(entry)
        return False



#!/usr/bin/env python
# -*- coding:utf-8 -*-

from __future__ import unicode_literals, print_function
import gevent
from gevent.queue import Queue

class Mail:
    def __init__(self, answer_box, message):
        self.answer_box = answer_box
        self.body = message

class Actor(gevent.Greenlet):
    def __init__(self):
        self.inbox = Queue()
        gevent.Greenlet.__init__(self)

    def on_message(self, message, answer_box=None):
        """ Define in your subclass. """
        raise NotImplemented()

    def send_to(self, dest, message):
        mail = Mail(self.inbox, message)
        if isinstance(dest, Actor):
            dest = dest.inbox
        dest.put(mail)

    def _run(self):
        self.running = True
        while self.running:
            message = self.inbox.get()
            if isinstance(message, Mail):
                self.on_message(message.body, answer_box=message.answer_box)
            else:
                self.on_message(message)
